#include <math.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <time.h>
#include "tron.h"

#ifndef min
template <class T> static inline T min(T x,T y) { return (x<y)?x:y; }
#endif

#ifndef max
template <class T> static inline T max(T x,T y) { return (x>y)?x:y; }
#endif

#ifdef __cplusplus
extern "C" {
#endif

extern double dnrm2_(int *, double *, int *);
extern double ddot_(int *, double *, int *, double *, int *);
extern int daxpy_(int *, double *, double *, int *, double *, int *);
extern int dscal_(int *, double *, double *, int *);

#ifdef __cplusplus
}
#endif

static void default_print(const char *buf)
{
	fputs(buf,stdout);
	fflush(stdout);
}

void TRON::info(const char *fmt,...)
{
	char buf[BUFSIZ];
	va_list ap;
	va_start(ap,fmt);
	vsprintf(buf,fmt,ap);
	va_end(ap);
	(*tron_print_string)(buf);
}

TRON::TRON(const function *fun_obj, double eps, double eps_cg, int max_iter)
{
	this->fun_obj=const_cast<function *>(fun_obj);
	this->eps=eps;
	this->eps_cg=eps_cg;
	this->max_iter=max_iter;
	tron_print_string = default_print;
}

TRON::~TRON()
{
}

void TRON::tron(double *w)
{
	clock_t start_time = clock();
	// Parameters for updating the iterates.
	double eta0 = 1e-4, eta1 = 0.25, eta2 = 0.75;

	// Parameters for updating the trust region size delta.
	double sigma1 = 0.25, sigma2 = 0.5, sigma3 = 4;

	int n = fun_obj->get_nr_variable();
	int i, cg_iter;
	double delta, snorm, one=1.0;
	double alpha, f, fnew, prered, actred, gs;
	int search = 1, iter = 1, inc = 1;
	double *s = new double[n];
	double *r = new double[n];
	double *g = new double[n];
	double *Ms = new double[n];

	// calculate gradient norm at w=0 for stopping condition.
	double *w0 = new double[n];
	for (i=0; i<n; i++)
		w0[i] = 0;
	fun_obj->fun(w0);
	fun_obj->grad(w0, g);
	double gnorm0 = dnrm2_(&n, g, &inc);
	delete [] w0;

	f = fun_obj->fun(w);
	fun_obj->grad(w, g);


	delta = dnrm2_(&n, g, &inc);
	double gnorm = delta;
	printf("delta origin = %lf, ", delta);

	if (gnorm <= eps*gnorm0)
		search = 0;

	//should be delta = gT M g
	delta = 0;
	double *Mg = new double[n];
	fun_obj->subHinvv(g, Mg, true);
	fun_obj->subHv(g, Mg);
	for (i=0;i<n;i++) delta += (g[i]*Mg[i]);
	delta = sqrt(delta);
	printf("delta now = %lf\n", delta);

	iter = 1;

	double *w_new = new double[n];
	bool reach_boundary;
	while (iter <= max_iter && search)
	{
		cg_iter = trcg(delta, g, s, r, &reach_boundary, Ms);
		snorm = sqrt(ddot_(&n, s, &inc, Ms, &inc));

		memcpy(w_new, w, sizeof(double)*n);
		daxpy_(&n, &one, s, &inc, w_new, &inc);

		gs = ddot_(&n, g, &inc, s, &inc);
		prered = -0.5*(gs-ddot_(&n, s, &inc, r, &inc));
		fnew = fun_obj->fun(w_new);

		// Compute the actual reduction.
		actred = f - fnew;

		// On the first iteration, adjust the initial step bound.
		//snorm = dnrm2_(&n, s, &inc);

		if (iter == 1)
			delta = min(delta, snorm);

		// Compute prediction alpha*snorm of the step.
		if (fnew - f - gs <= 0)
			alpha = sigma3;
		else
			alpha = max(sigma1, -0.5*(gs/(fnew - f - gs)));

		// Update the trust region bound according to the ratio of actual to predicted reduction.
		if (actred < eta0*prered)
			delta = min(max(alpha, sigma1)*snorm, sigma2*delta);
		else if (actred < eta1*prered)
			delta = max(sigma1*delta, min(alpha*snorm, sigma2*delta));
		else if (actred < eta2*prered)
			delta = max(sigma1*delta, min(alpha*snorm, sigma3*delta));
		else
		{
			if (reach_boundary)
				delta = sigma3*delta;
			else
				delta = max(delta, min(alpha*snorm, sigma3*delta));
		}

		info("iter %2d act %5.3e pre %5.3e delta %5.3e f %5.15e |g| %5.15e CG %3d time %5.6e\n", iter, actred, prered, delta, f, gnorm, cg_iter, float(clock()-start_time)/CLOCKS_PER_SEC);

		if (actred > eta0*prered)
		{
			iter++;
			memcpy(w, w_new, sizeof(double)*n);
			f = fnew;
			fun_obj->grad(w, g);

			gnorm = dnrm2_(&n, g, &inc);
			if (gnorm <= eps*gnorm0)
				break;
		}
		if (f < -1.0e+32)
		{
			info("WARNING: f < -1.0e+32\n");
			break;
		}
		if (prered <= 0)
		{
			info("WARNING: prered <= 0\n");
			break;
		}
		if (fabs(actred) <= 1.0e-12*fabs(f) &&
		    fabs(prered) <= 1.0e-12*fabs(f))
		{
			info("WARNING: actred and prered too small\n");
			break;
		}
	}

	delete[] g;
	delete[] r;
	delete[] w_new;
	delete[] s;
	delete[] Ms;
	delete[] Mg;
}

int TRON::trcg(double delta, double *g, double *s, double *r, bool *reach_boundary,
		double *Ms)
{
	int i, inc = 1;
	int n = fun_obj->get_nr_variable();
	double one = 1;
	double *d = new double[n];
	double *Hd = new double[n];
	//double rTr, rnewTrnew
	double alpha, beta, cgtol;
	double zTr, zTr_new;
	double *z = new double[n];
	double *Md = new double[n];
	//double *M_inv = new double[n];
	//fun_obj->getInvDiagH(M_inv);

	*reach_boundary = false;
	for (i=0; i<n; i++)
	{
		Ms[i] = s[i] = 0;
		r[i] = -g[i];
		//d[i] = r[i];
	}
	//cgtol = eps_cg*dnrm2_(&n, g, &inc);

	int cg_iter = 0;
	//rTr = ddot_(&n, r, &inc, r, &inc);
	fun_obj->subHinvv(r, z, true);
	//for (i=0; i<n; i++) z[i] = M_inv[i] * r[i];
	memcpy(d, z, sizeof(double)*(size_t)n);
	//memcpy(d, r, sizeof(double)*(size_t)n);
	memcpy(Md, r, sizeof(double)*(size_t)n);
	zTr = ddot_(&n, z, &inc, r, &inc);
	cgtol = eps_cg*sqrt(zTr);
	//printf("cgtol %5.3e\n", cgtol);
	while (1)
	{
		//if (dnrm2_(&n, r, &inc) <= cgtol)
		if (sqrt(zTr) <= cgtol)
			break;
		cg_iter++;
		fun_obj->Hv(d, Hd);
		//printf("%d ",cg_iter);
		alpha = zTr/ddot_(&n, d, &inc, Hd, &inc);
		//alpha = rTr/ddot_(&n, d, &inc, Hd, &inc);
		daxpy_(&n, &alpha, d, &inc, s, &inc);
		daxpy_(&n, &alpha, Md, &inc, Ms, &inc);
	
		//if (dnrm2_(&n, s, &inc) > delta)
		double sTMs = sqrt(ddot_(&n, Ms, &inc, s, &inc));
		//printf("sTMs %5.3e\n", sTMs);
		if (sTMs > delta)
		{
			info("cg reaches trust region boundary\n");
			*reach_boundary = true;
			alpha = -alpha;
			daxpy_(&n, &alpha, d, &inc, s, &inc);
			daxpy_(&n, &alpha, Md, &inc, Ms, &inc);

			double std = ddot_(&n, Ms, &inc, d, &inc);
			double sts = ddot_(&n, Ms, &inc, s, &inc);
			double dtd = ddot_(&n, Md, &inc, d, &inc);
			double dsq = delta*delta;
			double rad = sqrt(std*std + dtd*(dsq-sts));
			if (std >= 0)
				alpha = (dsq - sts)/(std + rad);
			else
				alpha = (rad - std)/dtd;
			daxpy_(&n, &alpha, d, &inc, s, &inc);
			daxpy_(&n, &alpha, Md, &inc, Ms, &inc);
			alpha = -alpha;
			daxpy_(&n, &alpha, Hd, &inc, r, &inc);
			break;
		}
		alpha = -alpha;
		daxpy_(&n, &alpha, Hd, &inc, r, &inc);

		fun_obj->subHinvv(r, z);
		//for (i=0; i<n; i++) z[i] = M_inv[i] * r[i];
		zTr_new = ddot_(&n, z, &inc, r, &inc);
		//rnewTrnew = ddot_(&n, r, &inc, r, &inc);
		beta = zTr_new/zTr;
		//beta = rnewTrnew/rTr;
		dscal_(&n, &beta, d, &inc);
		dscal_(&n, &beta, Md, &inc);
		//daxpy_(&n, &one, r, &inc, d, &inc);
		daxpy_(&n, &one, z, &inc, d, &inc);
		daxpy_(&n, &one, r, &inc, Md, &inc);
		//rTr = rnewTrnew;
		zTr = zTr_new;
	}

	delete[] d;
	delete[] Hd;
	delete[] z;
	delete[] Md;
	return(cg_iter);
}

double TRON::norm_inf(int n, double *x)
{
	double dmax = fabs(x[0]);
	for (int i=1; i<n; i++)
		if (fabs(x[i]) >= dmax)
			dmax = fabs(x[i]);
	return(dmax);
}

void TRON::set_print_string(void (*print_string) (const char *buf))
{
	tron_print_string = print_string;
}
